package rknop.jvmap;

import rknop.*;
import rknop.jvmap.*;
import java.lang.*;
import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;
import java.util.*;

// **********************************************************************
// **********************************************************************
// Counter keeps track of counters.  It keeps a pointer to the Image
//   of that counter (which is clipped from the big counter image, and
//   should be shared by all counters of the same type), as well as
//   the coordinates of the counter.  It has methods dealing with
//   moving.
// **********************************************************************
// **********************************************************************


class Counter extends Rectangle {
    static final private boolean debug=false;

    CounterSet cset;            // Counter Set that this is a memeber of
    CounterCanvas canvas;       // CounterCanvas we're on
    boolean drawn=false;        // Flag: should we draw it?
    public int x1,y1,x2,y2;     // Region into the counterSet to draw
    public int owner;           // Owner integer, read and saved
    public String comment;      // Comment string
    public int bitmap_index;
    int ordinal;                // Unique integer identifier of unit
    float facing;

    // **********************************************************************
    // constructor

    public Counter(CounterSet s) {
        super();
        canvas=null;
        cset=s;
        setSize(cset.counterwidth,cset.counterheight);
        x1=y1=x2=y2=0;
        drawn=false;
        comment="";
        owner=0;
        ordinal=0;
        facing=0;
    }
    public Counter(CounterSet s,int index,int left,int top) {
        super();

        drawn=false;
        canvas=null;
        cset=s;
        comment="";
        owner=0;
        ordinal=0;
        facing=0;

        setBitmapIndex(index);
        setBounds(left,top,cset.counterwidth,cset.counterheight);
    }
    public Counter(CounterSet s,int x,int y,int x1,int y1,int x2,int y2) {
        super();
        width=x2-x1+1;
        height=y2-y1+1;

        drawn=false;
        canvas=null;
        cset=s;
        comment="";
        owner=0;
        ordinal=0;
        facing=0;

        setBounds(x,y,width,height);
        setImage(x1,y1,x2,y2);
    }
    

    // **********************************************************************
    // methods to set the counter state (image, ancillary data)

    public void setImage(int x1,int y1,int x2,int y2) {
        this.x1=x1;
        this.y1=y1;
        this.x2=x2;
        this.y2=y2;

        setSize(x2-x1,y2-y1);
    
        bitmap_index=(y1/cset.counterheight)*cset.columns + 
            (x1/cset.counterwidth);

        // Undraw the old image
        boolean doredraw=drawn;
        undraw();
        if (doredraw) {
            draw();
            canvas.repaint(x,y,width,height);
        }
    }
    public void setBitmapIndex(int bi) {
        this.bitmap_index=bi;
        y1=cset.counterheight*(bi/cset.columns);
        x1=cset.counterwidth*(bi - cset.columns*(bi/cset.columns));
        y2=y1+cset.counterheight;
        x2=x1+cset.counterwidth;
        if (drawn) canvas.repaint(x,y,width,height);
    }
    public void setCSet(CounterSet s) { cset=s; }
    public void putOnCanvas(CounterCanvas c) { canvas=c; }
    public void setOrdinal(int n) { ordinal=n; }
    public void setOwner(int n) { owner=n; }
    public void setFacing(float f) { facing=f; }
    public void setComment(String s) { comment=new String(s); }


    // **********************************************************************
    // movecounter()
    //    Move to a new location, undrawing it at the old place and
    //       drawing it at the new.  NOTE -- coordinates are of the
    //       *center* of the counter

    public void moveCounter(int newx,int newy) {
        boolean wasdrawn=false;
        if ((canvas!=null) && drawn) { 
            wasdrawn=true;
            drawn=false;
            canvas.repaint(x,y,width,height);
        }
        x=newx-width/2;
        y=newy-width/2;
        if ((canvas!=null) && wasdrawn) { 
            drawn=true;
            canvas.repaint(x,y,width,height);
        }
    }


    // **********************************************************************
    // edit the comment

    class CounterCommentEdit implements ActionListener {
        Dialog dlg;
        Frame frm;
        TextField txt;
        Button ok,cancel;
        BorderLayout b;

        CounterCommentEdit(Component parent) {
            frm=Steege.getTopParent(parent);
            dlg=new Dialog(frm,"Edit Counter Comment",true);

            b=new BorderLayout();       // these two lines are gratuitous,
            dlg.setLayout(b);           //   remove them.

            txt=new TextField(Counter.this.comment);
            txt.addActionListener(this);
            dlg.add("North",txt);

            Panel p=new Panel();
            ok=new Button("OK");
            ok.setActionCommand("OK");
            ok.addActionListener(this);
            p.add(ok);
            cancel=new Button("Cancel");
            cancel.setActionCommand("Cancel");
            cancel.addActionListener(this);
            p.add(cancel);
            dlg.add("South",p);

            dlg.pack();
            dlg.show();
        }

        public void actionPerformed(ActionEvent e) {
            String com=e.getActionCommand();
            if (!com.equals("Cancel")) {
                Counter.this.comment=txt.getText();
            }
            dlg.removeAll();  // Without this before the dlg.hide(), 
                              //  and with a TextField,
                              //  it was hanging for 5 seconds on
                              //  the dlg.dispose() call.  (?)
            ok.removeActionListener(this);
            cancel.removeActionListener(this);
            txt.removeActionListener(this);
            dlg.hide();
            dlg.dispose();
        }

        public void finalize() throws Throwable {
            if (debug) {
                System.out.println("CounterCommentEdit finalize...");
            }
            super.finalize();
        }
    }

    public void editComment(Component parent) {
        CounterCommentEdit cedit=new CounterCommentEdit(parent);
        cedit=null;
        System.gc();
        System.runFinalization();
    }

    // **********************************************************************
    // Toggle the draw flags

    public void undraw() {
        drawn=false;
    }

    public void draw() {
        drawn=true;
    }


    // **********************************************************************
    // Render renders the counter on a given Graphics thingy.
    //
    // Note that render only does its stuff if the counter set is
    //  loaded.  We can be sure that the ImageObserver watching
    //  the counterset is going to ask to redraw the whole map when
    //  the thing finally loads, so we know this call will be made
    //  again before long if map.daddy.counterSetLoaded==false.
    //
    // Note 2 -- make this method accept a bounding rectangle!

    public void render(Graphics g,Rectangle bound) {
        if ((canvas!=null) && cset.counterSetLoaded) {
            int setx1,setx2,sety1,sety2;
            int imx1,imx2,imy1,imy2;

            setx1=x1;
            sety1=y1;

            if (bound.x>x) {
                if (bound.x>=x+width) return;
                imx1=bound.x;
                setx1+=bound.x-x;
            } else {
                imx1=x;
            }
            if (bound.y>y) {
                if (bound.y>=y+height) return;
                imy1=bound.y;
                sety1+=bound.y-y;
            } else {
                imy1=y;
            }
            if (bound.x+bound.width < x+width) {
                if (bound.x+bound.width <= x) return;
                imx2=bound.x+bound.width;
            } else {
                imx2=x+width;
            }
            if (bound.y+bound.height < y+height) {
                if (bound.y+bound.height <= y) return;
                imy2=bound.y+bound.height;
            } else {
                imy2=y+height;
            }
            setx2=setx1+imx2-imx1;
            sety2=sety1+imy2-imy1;

            g.drawImage(cset.image,imx1,imy1,imx2,imy2,
                        setx1,sety1,setx2,sety2,null);
        }
    }
}
