package rknop.jvmap;

import rknop.*;
import rknop.jvmap.*;
import java.lang.*;
import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;
import java.util.*;


// **********************************************************************
// **********************************************************************
// CounterCanvas is a canvas you can put coutners on.  It's got a list
//   of Counters that are on it, which it redraws as it feels necessary.
//
// **********************************************************************
// **********************************************************************

class CounterCanvas extends Canvas implements MouseListener,ActionListener
{
    static final private boolean debug=false;

    Image image;
    public boolean imageLoaded;
    Vector clist;                // Last element is on the bottom
    CounterSet cset;

    CounterPopup counterpopup;
    CounterPopupHandler counterhandler;
    PopupMenu bgpopup;

    Counter movingcounter;
    CounterDragger counterDragger;
    int pressx,pressy;


    // **********************************************************************
    // Constructor arguments:
    //
    // image= the background image, or null.

    // NOTE -- this currently doesn't support imagesless canvases
    //  (such as was previously used for the border).  There is some
    //  vestigal code for this, but it's not fully supported (no way
    //  of passing the width and height at the moment).
    

    public CounterCanvas(CounterSet cset, Image image) throws Exception {
        super();

        int width=0,height=0;

        this.cset=cset;
        this.image=image;

        clist=new Vector();

        if (image!=null) {
            Toolkit.getDefaultToolkit().prepareImage(image,-1,-1,this);
            MediaTracker mt=new MediaTracker(this);
            mt.addImage(image,0);
            try {
                mt.waitForAll();
            } catch(InterruptedException ex) {}
            
            if (mt.isErrorAny()) {
                throw new Exception("CounterCanvas: Error loading image!!");
            }

            width=image.getWidth(this);
            height=image.getHeight(this);
        }
        else {
            throw new Exception("CounterCanvas Error: null image.");
        }
        imageLoaded=true;
        
        setSize(width,height);

        // Create our popup menus:

        String[] labels;
        String[] commands;
        if (debug) {
            labels=new String[] { "Add Counter","Dump","Redraw" };
            commands=new String[] { "Add","Dump","Redraw" };
        } else {
            labels=new String[] { "Add Counter","Redraw" };
            commands=new String[] { "Add","Redraw" };
        }

        bgpopup=new PopupMenu();
        for (int i=0 ; i<labels.length ; ++i) {
            MenuItem mi=new MenuItem(labels[i]);
            mi.setActionCommand(commands[i]);
            mi.addActionListener(this);
            bgpopup.add(mi);
        }
        add(bgpopup);

        // Add the counter popup

        labels=new String[] { "Move","Flip","Change",
                             "Push to Bottom",
                             "Bring to Top","Delete",
                             "-Sep-","(Add Comment)"};
        commands=new String[] { "Move","Flip",
                                "Change","Bottom","Top",
                                "Delete","","Comment" };
        counterpopup=new CounterPopup();
        counterhandler=new CounterPopupHandler(this);
        for (int i=0 ; i<labels.length ; ++i) {
            if (labels[i].equals("-Sep-")) {
                counterpopup.addSeparator();
            } else {
                MenuItem mi=new MenuItem(labels[i]);
                mi.setActionCommand(commands[i]);
                mi.addActionListener(counterhandler);
                counterpopup.add(mi);
            }
        }
        add(counterpopup);

        addMouseListener(this);

        counterDragger=new CounterDragger();
        movingcounter=null;
    }

    // **********************************************************************
    // cleanup() -- destructor
    //
    // Note : don't clean up cset, as it might be used elsewhere.

    public void cleanup() {
        image=null;
        removeMouseListener(this);

        // Remove menu listeners

        for (int i=0 ; i<bgpopup.getItemCount() ; ++i) {
            bgpopup.getItem(i).removeActionListener(this);
        }
        for (int i=0 ; i<counterpopup.getItemCount() ; ++i) {
            counterpopup.getItem(i).removeActionListener(counterhandler);
        }

        // Remove popup menus
            
        remove(counterpopup);
        remove(bgpopup);

        // Remove counters explicitly, or trust them to be garbage collected?
    }

    public void finalize() throws Throwable {
        if (debug) {
            System.err.println("...in CounterCanvas finalizer");
        }
        super.finalize();
    }

    // **********************************************************************
    // counterList() returns the CounterList for this canvas

    public Vector counterList() {
        return clist;
    }


    // **********************************************************************
    // addCounter() adds a counter to the internal list of same
    //
    // NOTE!  It should not be on another list when you do this!

    public void addCounter(Counter c) {
        addCounter(c,false);
    }
    public void addCounter(Counter c,boolean toBottom) {
        if (toBottom) {
            clist.addElement(c);
        } else {
            clist.insertElementAt(c,0);
        }
        c.putOnCanvas(this);
        c.draw();
        repaint(c.x,c.y,c.width,c.height);
    }


    // **********************************************************************
    // removeCounter() removes a counter from the list

    public void removeCounter(Counter c) {
        clist.removeElement(c);
    }

    public void removeAllCounters() {
        clist=new Vector();
        repaint();
    }

    // **********************************************************************
    // flipAllCounters()

    public void flipAllCounters() {
        Counter c;
        Enumeration en=clist.elements();
        while (en.hasMoreElements()) {
            c=(Counter)en.nextElement();
            if (c.bitmap_index%2 > 0) {
                c.setBitmapIndex(c.bitmap_index-1);
            } else {
                c.setBitmapIndex(c.bitmap_index+1);
            }
        }
    }

    // **********************************************************************
    // setImage() -- sets a new map image

    public void setImage(Image newimage) {
        newimage=image;
        repaint();
    }

    // ****************************************
    // DragCounter starts a given counter dragging

    public void DragCounter(Counter which) {
        clist.removeElement(which);
        clist.insertElementAt(which,0);
        movingcounter=which;
        counterDragger.setLimits(this);
        addMouseMotionListener(counterDragger);
    }

    // **********************************************************************
    // update() just calls paint()

    public void update(Graphics g) {
        paint(g);
    }

    // **********************************************************************
    // paint() redraws the background map and all counters thereon

    public void paint(Graphics g) {
        Rectangle boundry=g.getClipBounds();
        int x0,x1,y0,y1,w,h;
        
        if (boundry!=null) {
            x0=boundry.x;
            y0=boundry.y;
            w=boundry.width;
            h=boundry.height;
            x1=x0+w;
            y1=y0+h;
        } else {
            Dimension d=getSize();
            x0=0;
            y0=0;
            w=d.width;
            x1=w;
            h=d.height;
            y1=h;
        }
        

        if (image!=null) {
            if (imageLoaded) {
                if (boundry==null) {
                    g.drawImage(image,0,0,this);
                } else {
                    g.drawImage(image,x0,y0,x1,y1,x0,y0,x1,y1,this);
                }
            }
        } else {
            g.setColor(Color.white);
            g.fillRect(x0,y0,w,h);
        }
        if (boundry!=null) {
            redrawcounters(g,boundry);
        } else {
            redrawcounters(g);
        }
    }
  
    // **********************************************************************
    // redrawcounters() causes all counters on the canvas to redraw.
    // 

    public void redrawcounters(Graphics g) {
        Rectangle r;
        Dimension d;
        int i;
        
        d=getSize();
        r=new Rectangle(0,0,d.width,d.height);
        redrawcounters(g,r);
    }
    
    public void redrawcounters(Graphics g,Rectangle boundry) {
        Counter c;
        int i;

        i=clist.size();
        for (--i ; i>=0 ; --i) {
            c=(Counter)(clist.elementAt(i));
            
            if (c.x <= boundry.x+boundry.width && c.x+c.width >= boundry.x &&
                c.y <= boundry.y+boundry.height && c.y+c.height >= boundry.y) {
                
                c.render(g,boundry);
            }
        }
        
    }

    

    // **********************************************************************
    // **********************************************************************
    // Event handlers
    
    // **********************************************************************
    // ActionPerformed() -- menu item on background screen called
    //   (part of the ActionListener interface)

    public void actionPerformed(ActionEvent e) {
        String command=e.getActionCommand();
        
        if (command.equals("Add")) {
            Counter c=cset.selectCounter(this);
            if (c!=null) {
                int ord=cset.registerOrdinal();
                c.setOrdinal(ord);
                c.moveCounter(pressx,pressy);
                addCounter(c);
            }
        }

        else if (command.equals("Redraw")) {
            // paint(getGraphics());
            // redrawcounters();                      // redundant?
            repaint();
        }
        else if (command.equals("Dump")) {
            int i=0;
            Counter c;

            System.gc();
            System.runFinalization();

            Enumeration enum=clist.elements();
            while (enum.hasMoreElements()) {
                c=(Counter)(enum.nextElement());
          
                System.out.print("Counter "+String.valueOf(i)+": ("+
                                 String.valueOf(c.x1)+"-"+String.valueOf(c.x2)+
                                 ") , ("+String.valueOf(c.y1)+"-"+
                                 String.valueOf(c.y2)+")  ;  "+
                                 "Position: "+String.valueOf(c.x)+","+
                                 String.valueOf(c.y)+"  ;  Bitmap Index: "+
                                 String.valueOf(c.bitmap_index)+"\n"
                                 );
                if (c.drawn) System.out.print(" ; drawn");
                else System.out.print(" ; NOT drawn\n");
                System.out.print(" ; width="+String.valueOf(c.width));
                System.out.print(" ; height="+String.valueOf(c.height)+"\n");
                ++i;
            }

        }
    }
    

    // **********************************************************************
    // CounterPopupHandler is a class that handles menu events from
    //   CounterCanvas.counterpopup

    class CounterPopupHandler implements ActionListener {

        public CounterPopupHandler(CounterCanvas daddy) { }

        public void actionPerformed(ActionEvent e) {
            String command=e.getActionCommand();
            Counter c;
            Rectangle r;

            if (counterpopup.counter()==null) return;

            c=counterpopup.counter();
            r=new Rectangle(c.x,c.y,c.width,c.height);

            if (command.equals("Bottom")) {
                removeCounter(c);
                addCounter(c,true);
                repaint(r.x,r.y,r.width,r.height);
            }
            else if (command.equals("Top")) {
                removeCounter(c);
                addCounter(c);
                repaint(r.x,r.y,r.width,r.height);
            }
            else if (command.equals("Delete")) {
                removeCounter(c);
                counterpopup.setCounter(null);
                repaint(r.x,r.y,r.width,r.height);
                // redrawcounters(r);                 // Redundant??
            }
            else if (command.equals("Change")) {
                Counter newc;

                newc=cset.selectCounter(CounterCanvas.this,
                                        "Select New Counter Image");
                if (newc!=null) {
                    c.setBitmapIndex(newc.bitmap_index);
                }
            }
            else if (command.equals("Flip")) {
                if (c.bitmap_index%2 > 0) {
                    c.setBitmapIndex(c.bitmap_index-1);
                } else {
                    c.setBitmapIndex(c.bitmap_index+1);
                }
            }
            else if (command.equals("Move")) {
                DragCounter(c);
            }
            else if (command.equals("Comment")) {
                c.editComment(CounterCanvas.this);
            }
        }
    }

    // ****************************************
    // CounterDragger lets you drag counters around with the mouse

    class CounterDragger implements MouseMotionListener {
        public int lx=0,ly=0;
        public int hx,hy;

        public void setLimits(Component c) {
            Dimension d=c.getSize();
            hx=d.width;
            hy=d.height;
        }

        public void mouseDragged(MouseEvent e) {
            mouseMoved(e);
        }
      
        public void mouseMoved(MouseEvent e) {
            if (movingcounter==null) {
                return;
            }
            int x=e.getX();
            int y=e.getY();
            // if (x<lx) x=lx;
            // if (x>hx) x=hx;
            // if (y<ly) y=ly;
            // if (y>hy) y=hy;
            movingcounter.moveCounter(x,y);
        }
    }

    // **********************************************************************
    // **********************************************************************
    // MouseListener methods

    
    // ********** Mouse Press

    public void mousePressed(MouseEvent e) {
        // First order of business -- if we're moving, ignore all presses
        if (movingcounter!=null) { return; }

        // Right.

        int mods=e.getModifiers();
        int x=e.getX();
        int y=e.getY();
        Counter c,oncounter;

        // Figure out if we're on a counter; stop on the first match
        oncounter=null;
        Enumeration enum=clist.elements();
        while (enum.hasMoreElements()) {
            c=(Counter)(enum.nextElement());
            if (c.contains(x,y)) {
                oncounter=c;
                break;
            }
        }
        
        if (oncounter==null) {                  // Pressed in empty space
            if (e.isPopupTrigger() || e.isShiftDown()) {
                pressx=x; pressy=y;
                if (bgpopup!=null) {
                    bgpopup.show(this,x,y);
                    e.consume();                   // (What is this for??)
                }
            }
        }         
        else {                                   // Pressed on a counter
            if (e.isPopupTrigger() || e.isShiftDown()) {

                // Modify the "comment" (last) item of counterpopup,
                //  and show it

                counterpopup.setCounter(oncounter);

                int n=counterpopup.getItemCount();
                MenuItem mi=counterpopup.getItem(n-1);

                if (oncounter.comment.length()>0) {
                    mi.setLabel(oncounter.comment);
                } else {
                    mi.setLabel("(Add Comment)");
                }
                counterpopup.show(this,x,y);
                e.consume();
            }
            else if ((mods & e.BUTTON2_MASK) >0 ) {
                
            }
            else if ((mods & e.BUTTON3_MASK) >0 ) {
                
            }
            else {                                      // Start a move
                DragCounter(oncounter);
            }
        }
    }
    
    // ********** Mouse Release -- for moving
    
    public void mouseReleased(MouseEvent e) {
        if (movingcounter!=null) {
            
            int x=e.getX();
            int y=e.getY();
            // YUKKKKKY use of counterDragger limits -- FIX!!!!!  ROB!!
            if (x<counterDragger.lx) x=counterDragger.lx;
            if (x>counterDragger.hx) x=counterDragger.hx;
            if (y<counterDragger.ly) y=counterDragger.ly;
            if (y>counterDragger.hy) y=counterDragger.hy;
            movingcounter.moveCounter(x,y);
            movingcounter=null;
            removeMouseMotionListener(counterDragger);
        }
    }



    // ********** Mouse Clicks
    
    public void mouseClicked(MouseEvent e) {
        int mods=e.getModifiers();
        int x=e.getX();
        int y=e.getY();
        Counter c,oncounter;
        
        // If we're moving, drop the guy we're moving and return
        
        if (movingcounter!=null) {
            // YUKKKKKY use of counterDragger limits -- FIX!!!!!  ROB!!
            if (x<counterDragger.lx) x=counterDragger.lx;
            if (x>counterDragger.hx) x=counterDragger.hx;
            if (y<counterDragger.ly) y=counterDragger.ly;
            if (y>counterDragger.hy) y=counterDragger.hy;
            movingcounter.moveCounter(x,y);
            movingcounter=null;
            removeMouseMotionListener(counterDragger);
            return;
        }
        
        // Figure out if we're on a counter (duplicate code :( )
        oncounter=null;
        Enumeration enum=clist.elements();
        while (enum.hasMoreElements()) {
            c=(Counter)(enum.nextElement());
            if (c.contains(x,y)) {
                oncounter=c;
                break;
            }
        }
        
        if (oncounter==null) {                // Clicked in empty space
            if (e.isPopupTrigger()) {
                pressx=x; pressy=y;
                if (bgpopup!=null) {
                    bgpopup.show(this,x,y);
                    e.consume();
                }
            }
        }
        else {                            // Clicked on a counter
            if (e.isPopupTrigger()) {
                counterpopup.setCounter(oncounter);
                counterpopup.show(this,x,y);
                e.consume();
            }
            else if ((mods & e.BUTTON2_MASK) > 0) {
                // Button 2 pushes it to the bottom if it's on top,
                //  otherwise brings it to the top
                if (  clist.indexOf(oncounter) == 0 ) {
                    clist.removeElementAt(0);
                    addCounter(oncounter,true);
                }
                else {
                    clist.removeElement(oncounter);
                    addCounter(oncounter);
                }
                return;
            }
        }
    }


    // ********** MouseEntered(), MouseExited() not used;

    public void mouseEntered(MouseEvent e) {
        return;
    }

    public void mouseExited(MouseEvent e) {
        return;
    }
}
