package rknop.jvmap;

import rknop.*;
import rknop.jvmap.*;
import java.lang.*;
import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;
import java.util.*;

// **********************************************************************
// **********************************************************************
// CounterSet is an image, divided into regular rectangles to 
//   represent square counters.
//
// NOTE -- I don't really want this to extend Component, but I seem
//   to be inable to use MediaTracker without it...!
// **********************************************************************
// **********************************************************************

class CounterSet extends Component {
    static final private boolean debug=false;
    
    String filename;

    public Image image;
    public boolean counterSetLoaded;
    
    public int counterheight,counterwidth,columns,rows;
    int imwidth,imheight;

    public int currentplayer;

    int pressx,pressy;            // Storage of mouse press location

    private Vector ordinalsused;

    // ****************************************
    // Constructor
    //
    // I *think* that it's very important that the image *not* be
    //   prepared... but I will test this.

    public CounterSet(Image inimage, int incols, int inrows) throws Exception 
    {
        ordinalsused=new Vector(20,20);
        
        imwidth=-1;
        imheight=-1;

        columns=incols;
        rows=inrows;

        counterSetLoaded=false;
        
        currentplayer=-1;

        image=inimage;
        Toolkit.getDefaultToolkit().prepareImage(image,-1,-1,this);
        MediaTracker mt=new MediaTracker(this);
        mt.addImage(image,0);
        mt.waitForAll();

        // Check errors!

        if (mt.isErrorAny()) {
            throw new Exception("Error loading counter set image!!!");
        }

        counterSetLoaded=true;

        imwidth=image.getWidth(this);
        imheight=image.getHeight(this);
        if ( (imwidth%columns !=0) || (imheight%rows !=0) ) {
            throw new Exception("Counter set doesn't "+
                                "divide evenly into counters!");
        }
        counterwidth=imwidth/columns;
        counterheight=imheight/rows;
    }

    // **********************************************************************
    // cleanup() -- destructor

    public void cleanup() {
        image=null;
    }

    public void finalize() throws Throwable {
        if (debug) {
            System.err.println("In CounterSet finalizer.");
        }
        super.finalize();
    }

    // **********************************************************************
    // setImage()
    //    change the image of the counter set.  The new image
    //    must have the same size as the old one, though, or we've
    //    got an exception on our hands

    public void setImage(Image newimage) throws Exception {
        Toolkit.getDefaultToolkit().prepareImage(newimage,-1,-1,this);
        MediaTracker mt=new MediaTracker(this);
        mt.addImage(newimage,0);
        mt.waitForAll();
        if (mt.isErrorAny()) {
            throw new Exception("Error loading image");
        }
        if (   newimage.getWidth(this)!=imwidth || 
               newimage.getHeight(this)!=imheight   ) {
            throw new Exception("Error: image size doesn't match size of "+
                                "current CounterSet image");
        }
        image=newimage;
        imwidth=newimage.getWidth(this);
        imheight=newimage.getHeight(this);
    }                                
        

    // **********************************************************************
    // registerOrdinal()
    //   Gets a new ordinal that isn't already used

    public synchronized int registerOrdinal() {
        if (ordinalsused.size()==0) {
            Integer i=new Integer(0);
            ordinalsused.addElement(i);
            return 0;
        }
        else {
            Integer last=(Integer)ordinalsused.lastElement();
            Integer i=new Integer(last.intValue()+1);
            ordinalsused.addElement(i);
            return i.intValue();
        }
    }

    public synchronized void registerOrdinal(int ord) throws Exception {
        Integer i,neword;

        if (ordinalsused.size()==0) {
            i=new Integer(ord);
            ordinalsused.addElement(i);
        }
        else {
            neword=new Integer(ord);

            for (int j=0 ; j<ordinalsused.size() ; ++j) {
                i=(Integer)ordinalsused.elementAt(j);
                if (ord==i.intValue()) {
                    throw new Exception("CounterSet Error: Duplicate Ordinal");
                }
                if (i.intValue()>ord) {
                    ordinalsused.insertElementAt(neword,j);
                    return;
                }
            }
            ordinalsused.addElement(neword);
        }
    }
                    
            


    // **********************************************************************
    // selectCounter() pops up a Dialog that lets you click and choose
    //   a counter.  When you create it, pass any component that's within
    //   the Frame that will form the Dialog's parent.

    private int selectCounterX,selectCounterY;
    private boolean selectCounterCancel;
    
    Counter selectCounter(Component parent) {
        return selectCounter(parent,"Select Counter...");
    }

    Counter selectCounter(Component parent,String title) {
        int x1,y1,x2,y2;
        CounterSelector selector;

        selector=new CounterSelector(parent,title);
        selector=null;                  // Clean up references
        System.gc();               // ****
        System.runFinalization();  // ****

        if (selectCounterCancel) return null;
            
        x1=selectCounterX-selectCounterX%counterwidth;
        y1=selectCounterY-selectCounterY%counterheight;
        x2=x1+counterwidth;
        y2=y1+counterheight;
        if ( (x1<0) || (y1<0) || 
             (x1>imwidth+1-counterwidth) || 
             (y1>imheight+1-counterheight) ) {
            return null;
        } else {
            Counter c=new Counter(this, 0, 0, x1, y1, x2, y2);
            if (currentplayer<=0) {
                c.setOwner(0);
            } else {
                c.setOwner(currentplayer);
            }
            return c;
        }
    }
    
    class CounterSelector extends MouseAdapter implements ActionListener
    {
        Dialog selectCounterDlg;
        CouSelCanvas canvas;
        //Canvas canvas;
        Button cancel;
        ScrollPane pane;

        class CouSelCanvas extends Canvas {
            CouSelCanvas() { super(); }
            public void paint(Graphics g) {
                g.drawImage(image,0,0,this);
            }
        
            public void finalize() throws Throwable {
                if (debug) {
                    System.err.println("...In CouSelCanvas finalizer");
                }
                super.finalize();
           }
        }

        public void actionPerformed(ActionEvent e) {
            String command=e.getActionCommand();
            if (command.equals("Cancel")) {
                selectCounterCancel=true;
                cleanup();
            }
        }

        public void mouseClicked(MouseEvent e) {
            selectCounterCancel=false;
            selectCounterX=e.getX();
            selectCounterY=e.getY();
            e.consume();
            cleanup();
        }

        public void cleanup() {
            if (debug) {
                System.err.println("Trying to cleanup CounterSelector");
            }
            selectCounterDlg.hide();
            selectCounterDlg.dispose();
            canvas.removeMouseListener(this);
            cancel.removeActionListener(this);
            pane.removeAll();
            selectCounterDlg.removeAll();
            canvas=null;
            pane=null;
            cancel=null;
            selectCounterDlg=null;
        }

        public CounterSelector(Component parent,String title) {
            Frame frm=Steege.getTopParent(parent);
            
            selectCounterDlg=new Dialog(frm,title,true);
            selectCounterDlg.setResizable(true);
            canvas=new CouSelCanvas();
            //canvas=new Canvas();
            canvas.setSize(imwidth,imheight);
            canvas.addMouseListener(this);
            pane=new ScrollPane();
            pane.setSize(400,300);
            pane.add(canvas);
            selectCounterDlg.add("Center",pane);
            
            cancel=new Button("Cancel");
            cancel.setActionCommand("Cancel");
            cancel.addActionListener(this);
            selectCounterDlg.add("South",cancel);
            
            selectCounterDlg.pack();

            // Center it on the CounterCanvas window...

            Point p=frm.getLocation();
            Dimension d=frm.getSize();
            Dimension s=selectCounterDlg.getSize();
            p.translate((d.width-s.width)/2,(d.height-s.height)/2);
            selectCounterDlg.setLocation(p);

            selectCounterDlg.show();
            if (debug) {
                System.err.println("Dude.");
            }
        }

        public void finalize() throws Throwable {
            if (debug) {
                System.err.println("...in CounterSelector finalizer");
            }
            super.finalize();
        }

    }
}    
