package rknop.jvmap;

import rknop.*;
import rknop.jvmap.*;
import java.lang.*;
import java.io.*;
import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;
import java.util.*;

// **********************************************************************
// **********************************************************************
// JVFrame is the class that opens up a Frame and puts a CounterCanvas
//   in it.
// **********************************************************************
// **********************************************************************

class JVFrame extends Frame implements ActionListener {
    static final private boolean debug=false;

    ScrollPane mapPane;
    CounterCanvas mapCanvas;
    String filename,title;
    int number,x0,y0,width,height;

    WinEventHndlr winlistener;

    MenuBar topMenu;
    MenuItem showmeitem;
    Vector showmelisteners;
    Vector menuitems;

    public JVFrame(  Image mapImage,CounterSet cset,
                     int number,int x0,int y0,int width,int height,
                     String filename,String title
                     ) throws Exception {
        super("JV_Map");

        Menu menu;
        MenuItem mi;
        int i,panewid,panehei;

        this.number=number;
        this.x0=x0;
        this.y0=y0;
        this.width=width;
        this.height=height;
        this.filename=new String(filename);
        this.title=new String(title);

        // Make sure the size of the image matches what we want
        // Note that I'm assuming the image is already loaded enough
        //   to get the width and height...!

        if (  (mapImage.getWidth(this)!=width && width!=0) ||
              (mapImage.getHeight(this)!=height && height!=0)) {
            throw new Exception("Error: Map image size is wrong!!!");
        }

        if (width==0) {
            width=mapImage.getWidth(this);
        }
        if (height==0) {
            height=mapImage.getHeight(this);
        }

        // Menu bar

        menuitems=new Vector(10,10);
        showmelisteners=new Vector(2,2);
        showmeitem=null;

        topMenu=new MenuBar();

        {
            String text[]={ "Flip All Units","Remove All Units","-Sep-",
                            "Show Main Window","Close" };
            String comm[]={ "Flip All","Remove All","-","Show Main","Close" };

            menu=new Menu("Tools");
            for (i=0 ; i<text.length ; ++i) {
                if (text[i].equals("-Sep-")) {
                    menu.addSeparator();
                } else {
                    mi=new MenuItem(text[i]);
                    mi.setActionCommand(comm[i]);
                    mi.addActionListener(this);
                    if (comm[i].equals("Show Main")) {
                        showmeitem=mi;
                    }
                    menuitems.addElement(mi);
                    menu.add(mi);
                }
            }
            topMenu.add(menu);
        }

        setMenuBar(topMenu);

        // CounterCanvas

        mapPane=new ScrollPane();
        add("Center",mapPane);

        mapCanvas=new CounterCanvas(cset,mapImage);
        mapPane.add(mapCanvas);

        // Windowlistener

        winlistener=new WinEventHndlr();
        addWindowListener(winlistener);

        pack();

        // Adjust size of mapPane and pack again

        if (debug) {
            System.err.println("About to reset size of mapPane");
        }

        panewid=550;
        panehei=650;
        if (width<panewid) panewid=width;
        if (height<panehei) panehei=height;
        panewid+=mapPane.getVScrollbarWidth();
        panehei+=mapPane.getHScrollbarHeight();
        mapPane.setSize(panewid,panehei);

        if (debug) {
            System.err.println("Set size of mapPane to "+
                               String.valueOf(panewid)+" , "+
                               String.valueOf(panehei));
        }

        pack();

        // show

        setVisible(true);
    }

    // **********************************************************************
    // addShowMeListener() -- adds a listener to the "Show Main" menu item

    public void addShowMeListener(ActionListener l) {
        showmeitem.addActionListener(l);
        showmelisteners.addElement(l);
    }

    // **********************************************************************
    // getCanvas() -- just in case we need to access the Counter Canvas
    //   (e.g. to add counters, read counters, etc.)

    public CounterCanvas getCanvas() {
        return mapCanvas;
    }

    // **********************************************************************
    // Window Event Handler

    private class WinEventHndlr extends WindowAdapter {
        public void windowClosing(WindowEvent e) {
            e.getWindow().setVisible(false);
        }
    }

    // **********************************************************************
    // Action Event Handler

    public void actionPerformed(ActionEvent e) {
        String com=e.getActionCommand();

        if (com.equals("Flip All")) {
            mapCanvas.flipAllCounters();
        }
        else if (com.equals("Remove All")) {
            String text[]={"Remove all units from map "+title+"?"};
            String buts[]={ "Remove","Cancel" };
            AreYouSure ays=new AreYouSure(this,"Remove Units",text,buts);
            if (ays.result.equals("Remove")) {
                mapCanvas.removeAllCounters();
            }
        }
        else if (com.equals("Close")) {
            setVisible(false);
        }
    }
                
    // **********************************************************************
    // cleanup()

    public void cleanup() {
        Enumeration en;
        ActionListener l;
        MenuItem m;

        setVisible(false);
        mapCanvas.cleanup();
        removeWindowListener(winlistener);

        en=menuitems.elements();
        while (en.hasMoreElements()) {
            m=(MenuItem)en.nextElement();
            m.removeActionListener(this);
        }

        en=showmelisteners.elements();
        while (en.hasMoreElements()) {
            l=(ActionListener)en.nextElement();
            showmeitem.removeActionListener(l);
        }

        removeAll();
        dispose();
    }

}
