package rknop.jvmap;

import rknop.*;
import rknop.jvmap.*;
import java.lang.*;
import java.io.*;
import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;
import java.util.*;

// **********************************************************************
// **********************************************************************
// JV_Map is a E-mail PBEM application.  The main class opens up a 
//   Window that lists the maps and players of this game, and then
//   lets you toggle open and closed a window for each map.
//
// **********************************************************************
// **********************************************************************

class JV_Map extends Frame implements ActionListener {
    static final private boolean debug=false;

    public static String jvmapversion="0.10.8";

    Vector maps;            // A vector of JVFrame
    Vector players;         // A vector of PlayerInfo

    CounterSet cset;

    String gametitle;
    String unparsedlines[];

    VMPFile vmpfile;
    String currentdirectory;

    MenuBar topMenu;
    Button buttons[];
    Label tilelabel;
    Label titlelabel;
    Label dirlabel;
    Label filelabel;
    Label playerlabel;
    Vector menuitems;
    java.awt.List maplist,playerlist;

    WinEventHndlr winlistener;

    // **********************************************************************
    // Constructor

    public JV_Map() throws Exception {
        super("JV_Map "+jvmapversion);
        construct(null);
    }

    public JV_Map(String vmpfilename) throws Exception {
        super("JV_Map "+jvmapversion);
        File file=new File(vmpfilename);
        VMPFile vmp=new VMPFile();
        
        if (debug) {
            System.err.println("About to read "+vmpfilename+" from "+vmp.dir);
        }

        vmp.readVMPFile(vmpfilename);
        if (vmp!=null) {
            String error=vmp.errorMsg();
            if (error.length()>0) {
                System.err.println("NOTE:");
                System.err.println(error);
            }

            if (debug) {
                System.err.println("Read "+vmpfilename+" from "+vmp.dir);
            }

            File f=new File(vmpfilename);
            File d=new File(f.getParent());
            currentdirectory=d.getAbsolutePath();
            construct(vmp);
        } else {
            System.err.println("Error reading VMP file "+vmpfilename);
            System.exit(10);
        }
    }

    // **********************************************************************
    // construct() -- Actually initialize and build the UI
    //   You know, life would be a lot easier of Java 1.1 had the BoxLayout.
    //   I mean, sheesh, it's the most basic Layout you can imagine, and I'm
    //   pretty sure Motif has it, since IDL does.  Instead, we're stuck with
    //   the very limited Flow and Grid layouts.  Border is fun, but also
    //   limited.  And I refuse to deal with GridBag.  I just want
    //   BoxLayout!!!!!!  Instead, I nest lots of Borders and Grids.  Yeah, 
    //   I probably ought to figure out GridBag layout, but it just seems
    //   a bit much.

    private void construct(VMPFile vmpfile) throws Exception {
        int width,height,i,j;
        MenuBar topMenu;
        Menu menu;
        MenuItem mi;
        ScrollPane sp;
        Panel mainpanel,pnl,p;
        Label l;
        BorderLayout blo;
        FlowLayout flo;
        GridLayout glo;

        Font bigSerif=new Font("Serif",Font.BOLD,14);
        Font boldnormal=new Font("SansSerif",Font.BOLD,12);
        Font plainnormal=new Font("SansSerif",Font.PLAIN,12);

        if (debug) {
            System.err.println("In JV_Map constructor");
        }

        String[] mapbuttonText={"Show","Hide","Show All",
                                "Hide All","Remove","Edit",
                                "Add"};
        String[] mapbuttonCmd={"Show Map","Hide Map","Show All Maps",
                            "Hide All Maps","Remove Map","Edit Map",
                            "Add Map"};
        String[] plbuttonText={"Add","Edit","Remove"};
        String[] plbuttonCmd={"Add Player","Edit Player","Remove Player"};

        buttons=new Button[mapbuttonText.length+plbuttonText.length+1];

        // Initialize

        System.gc();                   // Gratuitous anal retensivity
        System.runFinalization();
        
        cset=null;
        maps=new Vector(1,3);
        players=new Vector(2,3);
        menuitems=new Vector(20,10);
        winlistener=new WinEventHndlr();
        addWindowListener(winlistener);

        // Read our config file

        File configfile=new File(System.getProperty("user.home"),".jvmaprc");
        if (configfile.exists() && configfile.canRead()) {
            Properties props=new Properties(System.getProperties());
            props.load(new BufferedInputStream
                (new FileInputStream(configfile.getPath())));
            System.setProperties(props);
        }

        if (debug) {
            String smtpserver=System.getProperty("rknop.jvmap.smtpserver");
            if (smtpserver!=null) {
                System.err.println("SMTP Server: "+smtpserver);
            }
        }

        // Build the GUI

        try {

            // Make the main menubar
            
            topMenu=new MenuBar();

            {
                String menuText[]={"New","Load","Save VMP 1.5",
                                   "Save VMP 2.0","-Sep-","Exit"};
                String menuCmd[]={"New VMP","Load VMP","Save VMP 1.5",
                                  "Save VMP 2.0","","Exit JV_Map"};
                menu=menuBuilder("File",menuText,menuCmd,this);
                topMenu.add(menu);
            }

            {
                String menuText[]={"Edit Game Title","Roll Dice"};
                String menuCmd[]={"Edit Game Title","Roll Dice"};
                menu=menuBuilder("Tools",menuText,menuCmd,this);
                topMenu.add(menu);
            }

            {
                String menuText[]={"About JV_Map"};
                String menuCmd[]={"About JV_Map"};
                menu=menuBuilder("Help",menuText,menuCmd,this);
                topMenu.setHelpMenu(menu);
            }

            setMenuBar(topMenu);

            // Info at the top

            pnl=new Panel();
            glo=new GridLayout(4,1);
            pnl.setLayout(glo);
            add("North",pnl);

            p=new Panel();
            blo=new BorderLayout();
            p.setLayout(blo);
            l=new Label("Game: ");
            l.setFont(boldnormal);
            p.add("West",l);
            titlelabel=new Label("");
            titlelabel.setFont(plainnormal);
            p.add("Center",titlelabel);
            pnl.add(p);

            p=new Panel();
            blo=new BorderLayout();
            p.setLayout(blo);
            l=new Label("Player: ");
            l.setFont(boldnormal);
            p.add("West",l);
            playerlabel=new Label("");
            playerlabel.setFont(plainnormal);
            p.add("Center",playerlabel);
            pnl.add(p);

            p=new Panel();
            blo=new BorderLayout();
            p.setLayout(blo);
            l=new Label("Filename: ");
            l.setFont(boldnormal);
            p.add("West",l);
            filelabel=new Label("");
            filelabel.setFont(plainnormal);
            p.add("Center",filelabel);
            pnl.add(p);

            p=new Panel();
            blo=new BorderLayout();
            p.setLayout(blo);
            l=new Label("Directory: ");
            l.setFont(boldnormal);
            p.add("West",l);
            dirlabel=new Label("");
            dirlabel.setFont(plainnormal);
            p.add("Center",dirlabel);
            pnl.add(p);
            

            mainpanel=new Panel();
            blo=new BorderLayout();
            mainpanel.setLayout(blo);
            add("Center",mainpanel);


            // List of maps, and buttons beneath to control them
            
            pnl=new Panel();
            mainpanel.add("North",pnl);
            blo=new BorderLayout();
            pnl.setLayout(blo);
            
            p=new Panel();
            pnl.add("North",p);
            l=new Label("Maps");
            l.setFont(bigSerif);
            p.add(l);

            maplist=new java.awt.List();
            maplist.setMultipleMode(false);
            pnl.add("Center",maplist);
            
            p=new Panel();
            flo=new FlowLayout();
            p.setLayout(flo);
            pnl.add("South",p);
            
            for (i=0 ; i<mapbuttonText.length ; ++i) {
                buttons[i]=new Button(mapbuttonText[i]);
                buttons[i].setActionCommand(mapbuttonCmd[i]);
                buttons[i].addActionListener(this);
                p.add(buttons[i]);
            }
            

            // List of players, and buttons
            
            pnl=new Panel();
            mainpanel.add("Center",pnl);
            blo=new BorderLayout();
            pnl.setLayout(blo);
            
            p=new Panel();
            pnl.add("North",p);
            l=new Label("Players");
            l.setFont(bigSerif);
            p.add(l);

            playerlist=new java.awt.List();
            playerlist.setMultipleMode(false);
            pnl.add("Center",playerlist);
            
            p=new Panel();
            flo=new FlowLayout();
            p.setLayout(flo);
            pnl.add("South",p);
            
            for (i=0 ; i<plbuttonText.length ; ++i) {
                j=i+mapbuttonText.length;
                buttons[j]=new Button(plbuttonText[i]);
                buttons[j].setActionCommand(plbuttonCmd[i]);
                buttons[j].addActionListener(this);
                p.add(buttons[j]);
            }
            
            // Counter set gadget deal thingy
            
            p=new Panel();
            blo=new BorderLayout();
            p.setLayout(blo);
            mainpanel.add("South",p);
            
            j=mapbuttonText.length + plbuttonText.length;
            buttons[j]=new Button("Change");
            buttons[j].setActionCommand("Change Tiles");
            buttons[j].addActionListener(this);
            p.add("West",buttons[j]);
            
            tilelabel=new Label("Tiles image: (none)");
            p.add("Center",tilelabel);
            
            
            // Show the main window
            
            pack();
            setVisible(true);

            if (vmpfile!=null) {
                this.vmpfile=vmpfile;
                buildMeFromVMP(vmpfile);
            }
            else {
                this.vmpfile=new VMPFile();
                File f1=new File(System.getProperty("user.dir"));
                File f2=new File(f1.getAbsolutePath());
                String dir=f2.getParent();
                if (dir!=null) {
                    this.vmpfile.dir=new String(dir);
                }
                dirlabel.setText(this.vmpfile.dir);
            }
        }
        catch(Exception ex) {
            cleanup();
            throw ex;            // Does this work?
        }
    }

    // **********************************************************************
    // **********************************************************************
    // buildMeFromVMP()
    //    Pass a vmpfile, and the state of this object will be set
    //     based on its contents.  NOTE that clear() is assumed to have
    //     been run before this is called!

    private void buildMeFromVMP(VMPFile vmp) {
        int i,j;

        if (debug) {
            System.err.println("In buildMeFromVMP");
        }

        vmpfile=vmp;
        gametitle=new String(vmpfile.title);
        setTitle("JV_Map "+jvmapversion+": "+gametitle);
        maps=new Vector(vmpfile.maps.length,3);
        players=new Vector(vmpfile.players.length,3);
            
        titlelabel.setText(gametitle);
        filelabel.setText(vmpfile.filename);
        dirlabel.setText(vmpfile.dir);
        currentdirectory=vmpfile.dir;
        
        try {

            // **************** Create the counter set
            
            {
                Image image;
                
                try {
                    image=getImage(vmpfile.dir,vmpfile.tiles.filename);
                }
                catch (Exception ex) {
                    image=null;
                }
                if (image!=null) {
                    cset=new CounterSet(image,vmpfile.tiles.columns,
                                        vmpfile.tiles.rows);
                    cset.filename=new String(vmpfile.tiles.filename);
                    tilelabel.setText("Tiles image: "+cset.filename+" ("+
                                      String.valueOf(cset.columns)+" x "+
                                      String.valueOf(cset.rows)+")");
                }
                else {
                    String text[]={ "Failed to get Tiles file "+
                                    vmpfile.tiles.filename,
                                    "from directory "+vmpfile.dir };
                    String buttons[]={ "Abort." };
                    AreYouSure oops=new AreYouSure(this,"JV_Map Error",
                                                   text,buttons);
                    clear(false);
                    return;
                }

            }


            // *************** Create our maps

            if (debug) {
                System.err.println("Building Maps");
            }
            
            for (i=0 ; i<vmpfile.maps.length ; ++i) {
                MapInfo map=vmpfile.maps[i];
            
                Image image=getImage(vmpfile.dir,map.filename);
                
                JVFrame jvf=new JVFrame(image,cset,map.number,
                                        map.x0,map.y0,map.width,map.height,
                                        map.filename,map.title);

                // Populate it

                for (j=0 ; j<vmpfile.units.length ; ++j) {
                    if (vmpfile.units[j].map==vmpfile.maps[i].number) {
                        UnitInfo u=vmpfile.units[j];
                        Counter c=new Counter(cset,u.index,u.left,u.top);
                        c.setOwner(u.owner);
                        c.setFacing(u.facing);
                        cset.registerOrdinal(u.ordinal);
                        c.setOrdinal(u.ordinal);
                        c.setComment(u.comment);
                        jvf.getCanvas().addCounter(c,true);
                    }
                }

                jvf.setTitle("JV_Map "+jvmapversion+": "+jvf.title);
                jvf.addWindowListener(winlistener);
                jvf.addShowMeListener(this);
                synchronized (maps) {
                    maps.addElement(jvf);
                }

                // Add to the map list gadget
            
                String temp=new String("Map "+
                                       String.valueOf(vmpfile.maps[i].number)+
                                       ": "+vmpfile.maps[i].title+" ("+
                                       vmpfile.maps[i].filename+")");
                maplist.add(temp);

            }

            // *************** Create our players
        
            for (i=0 ; i<vmpfile.players.length ; ++i) {
                PlayerInfo player=new PlayerInfo();
                player.number=vmpfile.players[i].number;
                player.email=new String(vmpfile.players[i].email);
                player.password=new String(vmpfile.players[i].password);
                player.name=new String(vmpfile.players[i].name);
                synchronized (players) {
                    players.addElement(player);
                }
            
                // Add to the playerlist gadget
            
                String temp=new String("Player "+
                                       String.valueOf(player.number)+
                                       ": "+player.name+" <"+
                                       player.email+">");
                playerlist.add(temp);
            }            


            // Finally, save any unparsed lines, for later writing out...

            unparsedlines=vmpfile.other;

        }
        catch (Exception ex) {
            
            // DO SOMETHING INTELLIGENT
            
        }

        pack();        // Make sure everything is updated
    }

    // **********************************************************************
    // syncVMPFile() -- Turn the vmpfile into something which reflects the
    //                    current state.

    public void syncVMPFile() {
        JVFrame jvf;
        PlayerInfo player;
        int i,j,uptr,numunits;
        Vector clist;
        Counter c;

        if (cset==null) {
            System.err.println("Doh!!!  syncVMPfile failure.");
            System.exit(0);
        }

        VMPFile vmp=new VMPFile();

        vmp.dir=new String(vmpfile.dir);
        vmp.filename=new String(vmpfile.filename);
        vmp.title=new String(gametitle);
        vmp.vmpnum=vmpfile.vmpnum;

        numunits=0;
        vmp.maps=new MapInfo[maps.size()];
        for (i=0 ; i<maps.size() ; ++i) {
            jvf=(JVFrame)maps.elementAt(i);
            vmp.maps[i]=new MapInfo();
            vmp.maps[i].number=jvf.number;
            vmp.maps[i].x0=jvf.x0;
            vmp.maps[i].y0=jvf.y0;
            vmp.maps[i].width=jvf.width;
            vmp.maps[i].height=jvf.height;
            vmp.maps[i].filename=new String(jvf.filename);
            vmp.maps[i].title=new String(jvf.title);
            numunits+=jvf.getCanvas().counterList().size();
        }

        vmp.tiles.filename=new String(cset.filename);
        vmp.tiles.rows=cset.rows;
        vmp.tiles.columns=cset.columns;

        vmp.players=new PlayerInfo[players.size()];
        for (i=0 ; i<players.size() ; ++i) {
            player=(PlayerInfo)players.elementAt(i);
            vmp.players[i]=new PlayerInfo();
            vmp.players[i].number=player.number;
            vmp.players[i].email=new String(player.email);
            vmp.players[i].password=new String(player.password);
            vmp.players[i].name=new String(player.name);
        }

        vmp.units=new UnitInfo[numunits];
        uptr=0;
        for (i=0 ; i<maps.size() ; ++i) {
            jvf=(JVFrame)maps.elementAt(i);
            clist=jvf.getCanvas().counterList();
            for (j=0 ; j<clist.size() ; ++j,++uptr) {
                c=(Counter)clist.elementAt(j);
                vmp.units[uptr]=new UnitInfo();
                vmp.units[uptr].ordinal=c.ordinal;
                vmp.units[uptr].top=c.y;
                vmp.units[uptr].left=c.x;
                vmp.units[uptr].map=jvf.number;
                vmp.units[uptr].index=c.bitmap_index;
                vmp.units[uptr].owner=c.owner;
                vmp.units[uptr].facing=c.facing;
                vmp.units[uptr].comment=new String(c.comment);
            }
        }

        vmp.other=new String[unparsedlines.length];
        for (i=0 ; i<unparsedlines.length ; ++i) {
            vmp.other[i]=new String(unparsedlines[i]);
        }

        // Lines, rulers
        
        // Finally, make vmpfile the new thing

        vmpfile=vmp;
    }

        


    // **********************************************************************
    // **********************************************************************
    // ActionListener event handler
    //
    //    It's really ugly to have this all in one function.  Ideally,
    //       I'd have lots of little classes which are each ActionLiseners,
    //       each of which define only the one method.  This would be like
    //       having a different callback for each button.  However, then
    //       I'd have to keep track of all of them in order to be able to
    //       remove the listeners when I need to clean up... and since that
    //       would be a pain, we have a switchboard function.

    public void actionPerformed(ActionEvent e) {
        String command=e.getActionCommand();
        int whichmap,whichplayer;
        JVFrame jvf;
        PlayerInfo pl;
        synchronized (maplist) {
            whichmap=maplist.getSelectedIndex();
            if (whichmap>-1) {
                jvf=(JVFrame)maps.elementAt(whichmap);
            } else {
                jvf=null;
            }
        }
        synchronized (playerlist) {
            whichplayer=playerlist.getSelectedIndex();
            if (whichplayer>-1) {
                pl=(PlayerInfo)players.elementAt(whichplayer);
            } else {
                pl=null;
            }
        }
        

        // ************** Top Menu events

        if (command.equals("About JV_Map")) {
            AboutBox ab=new AboutBox(this);
        }
        else if (command.equals("New VMP")) {
            clear(true);
            getNewVMPName();
        }
        else if (command.equals("Load VMP")) {
            if (clear(cset!=null)) {
                VMPFile vmp=new VMPFile();
                vmp.dir=currentdirectory;
                if (!vmp.loadVMPFile(this)) {
                    String text[]={"Error reading VMP file:",vmp.errorMsg() };
                    String buts[]={ "Damn" };
                    AreYouSure ays=new AreYouSure(this,"VMP Error",text,buts);
                    return;
                }
                buildMeFromVMP(vmp);
            }
        }
        else if (command.equals("Save VMP 1.5")) {
            saveVMP15();
        }
        else if (command.equals("Save VMP 2.0")) {
            saveVMP20();
        }
        else if (command.equals("Exit JV_Map")) {
            checkClose();
        }

        else if (command.equals("Edit Game Title")) {
            TitleEditor te=new TitleEditor();
        }
        else if (command.equals("Roll Dice")) {
            DiceMail dm=new DiceMail(players);
        }

        // ****************** Map Buttons

        else if (command.equals("Show Map")) {
            if (whichmap>-1) {
                jvf.setVisible(true);
            }
        }
        else if (command.equals("Hide Map")) {
            if (whichmap>-1) {
                jvf.setVisible(false);
            }
        }
        else if (command.equals("Show All Maps")) {
            JVFrame map;
            Enumeration en=maps.elements();
            while (en.hasMoreElements()) {
                map=(JVFrame)en.nextElement();
                map.setVisible(true);
            }
        }
        else if (command.equals("Hide All Maps")) {
            JVFrame map;
            Enumeration en=maps.elements();
            while (en.hasMoreElements()) {
                map=(JVFrame)en.nextElement();
                map.setVisible(false);
            }
        }
        else if (command.equals("Remove Map")) {
            if (whichmap>-1) removeMap(jvf);
        }
        else if (command.equals("Edit Map")) {
            if (whichmap>-1) {
                MapEditor me=new MapEditor(jvf);
            }
        }
        else if (command.equals("Add Map")) {
            MapEditor me=new MapEditor(null);
        }


        // ******************** Player buttons

        else if (command.equals("Add Player")) {
            PlayerEditor pe=new PlayerEditor(null);
        }
        else if (command.equals("Edit Player")) {
            if (whichplayer>-1) {
                PlayerEditor pe=new PlayerEditor(pl);
            }
        }
        else if (command.equals("Remove Player")) {
            if (whichplayer>-1) removePlayer(pl);
        }


        // *********************** Tiles button

        else if (command.equals("Change Tiles")) {
            CSetEditor cse=new CSetEditor();
        }

        // *********************** ShowMe deals

        else if (command.equals("Show Main")) {
            setVisible(true);
        }

        // ************************ Dude

        else {
            if (debug) {
                System.err.println("Unknown command: "+command);
            }
        }

    }

    // **********************************************************************
    // notImpl() -- a function which *is* implemented

    public void notImpl() {
        String text[]={"Function not implemented."};
        String buttons[]={"Oh Well"};
        AreYouSure oops=new AreYouSure(this,"JV_Map Error",text,buttons);
    }


    // **********************************************************************
    // **********************************************************************
    // Our WindowListener

    private class WinEventHndlr extends WindowAdapter {
    
        // ******************************************************************
        // windowClosing() -- If from us, hide if there are maps visible,
        //                    otherwise quit.  If from a map, show us
        //                    if there are no maps visible.


        public void windowClosing(WindowEvent e) {
            boolean anymapsvisible=false,isus=true;
            Enumeration en;

            // If it wasn't us that generated the event, make sure the
            //   map window is closed.
            
            if (e.getWindow()!=(Window)JV_Map.this) {
                e.getWindow().setVisible(false);
                isus=false;
            }

            // See if there are any open maps
        
            en=maps.elements();
            while (en.hasMoreElements()) {
                JVFrame map=(JVFrame)en.nextElement();
                if (map.isVisible()) {
                    anymapsvisible=true;
                    break;
                }
            }


            if (anymapsvisible) {
                if (isus) setVisible(false);
            } else {
                if (isus) checkClose();
                else setVisible(true);
            }
        }
    }

    // **********************************************************************
    // clear() -- clears out everything

    synchronized boolean clear(boolean check) {
        if (check) {
            String text[]={ "Close current game:","Are you sure?" };
            String buttons[]= { "Close","Cancel" };
            AreYouSure db=new AreYouSure(this,"Close",text,buttons);
            if (db.result.equals("Cancel")) {
                return false;
            }
        }
        
        JVFrame jvf;
        Enumeration en=maps.elements();
        while (en.hasMoreElements()) {
            jvf=(JVFrame)en.nextElement();
            jvf.setVisible(false);
            jvf.cleanup();
        }
        if (cset!=null) { cset.cleanup(); }
        cset=null;
        gametitle=new String("");
        maps=new Vector(1,3);
        players=new Vector(1,3);
        maplist.removeAll();
        playerlist.removeAll();
        unparsedlines=new String[0];
        titlelabel.setText("");
        dirlabel.setText("");
        filelabel.setText("");
        tilelabel.setText("Tiles image: (none)");
        vmpfile.clear();

        setTitle("JV_Map "+jvmapversion);
        pack();

        System.gc();                 // Gratuitous cleanup
        System.runFinalization();    // because I'm anal retentive
        System.gc();                 // at heart

        return true;
    }


    // **********************************************************************
    // cleanup() -- our destructor

    void cleanup() {
        clear(false);
        
        // TODO: clean up our own act!!!!
    }

    // **********************************************************************
    // checkClose() makes sure you want to close the thing you just closed

    void checkClose() {
        String text[]={ "Quit JV_Map:","Are you sure?" };
        String buttons[]={ "Quit","Cancel" };
        AreYouSure db=new AreYouSure(this,"Close",text,buttons);
        if (db.result.equals("Quit")) {
            cleanup();
            System.exit(0);
        }
    }

    // **********************************************************************
    // finalize() -- called upon garbage colleciton

    protected void finalize() throws Throwable {
        if (debug) {
            System.out.println("Finalizing a JV_Map");
        }
        super.finalize();
    }

    // **********************************************************************
    // getImage() -- given an image filename, load the image, and
    //                make sure the bloody thing is loaded.
    //                (Yeah, yeah, I know, I'm supposed to take advantage
    //                of Java's multithreadedness, but it gets to be
    //                a pain to have to keep implementing ImageObserver
    //                methods just to make sure I know the width and
    //                height of the damn image.)

    Image getImage(String dir,String filename) throws Exception {
        Image image;

        File file=new File(dir,filename);

        image=BMPLoader.load(file.getPath());
        if (image==null) {
            if (debug) {
                System.err.println("Fail BMP for image, trying native Java.");
            }
            if (!file.exists() || !file.canRead()) {
                throw new Exception("Can't read image file "+file.getPath());
            }
            image=Toolkit.getDefaultToolkit().getImage(file.getPath());
            if (image==null) {
                throw new Exception("Can't read image file "+file.getPath());
            }
        }

        Toolkit.getDefaultToolkit().prepareImage(image,-1,-1,this);
        MediaTracker mt=new MediaTracker(this);
        mt.addImage(image,0);
        try {
            mt.waitForAll();
        } catch (InterruptedException ex) {
            throw new Exception("Interrupt Exception while loading image "+
                                file.getPath());
        }

        if (mt.isErrorAny()) {
            throw new Exception("Error loading image "+file.getPath());
        }

        return image;
    }


    // **********************************************************************
    // menuBuilder() takes a list of labels and commands, and an
    //   ActionLisener, and returns a menu suitable for adding to a 
    //   MenuBar

    private Menu menuBuilder(String name,String labels[],String commands[],
                             ActionListener listener) {
        Menu menu;
        MenuItem mi;
        int i;

        menu=new Menu(name);
        for (i=0 ; i<labels.length ; ++i) {
            if (labels[i].equals("-Sep-")) {
                menu.addSeparator();
            } else {
                mi=new MenuItem(labels[i]);
                mi.setActionCommand(commands[i]);
                mi.addActionListener(listener);
                menu.add(mi);
            }
        }
        return menu;
    }

    // **********************************************************************
    // getNewVMPName() prompts the user for a directory, and stuffs that
    //   in vmpfile.dir

    private void getNewVMPName() {
        FileDialog filebox=new FileDialog(this,"Select Directory",
                                          FileDialog.SAVE);
        filebox.setDirectory(currentdirectory);
        filebox.setVisible(true);
        String name=filebox.getFile();

        if (name==null) { return; }

        String dir=filebox.getDirectory();
        vmpfile.dir=new String(dir);
        currentdirectory=vmpfile.dir;
        vmpfile.filename=new String(name);
        filelabel.setText(vmpfile.filename);
        dirlabel.setText(vmpfile.dir);
    }        

    // **********************************************************************
    // saveVMP15() -- write a VMP version 1.5 file

    private void saveVMP15() {
        syncVMPFile();
        String ofname=getSaveFilename("Write VMP Version 1.5 file",vmpfile.dir,
                                      vmpfile.filename);
        if (ofname==null) return;

        try {
            vmpfile.writeVMP15(ofname);

            filelabel.setText(vmpfile.filename);
            dirlabel.setText(vmpfile.dir);
            currentdirectory=vmpfile.dir;

        } catch (Exception ex) {
            String text[]={"Error writing file "+ofname,""};
            String buts[]={ "Nuts" };
            text[1]=ex.getMessage();
            AreYouSure ays=new AreYouSure(this,"Error Writing VMP 1.5",
                                          text,buts);
        }
    }

    // **********************************************************************
    // saveVMP20() -- write a VMP version 2.0 file

    private void saveVMP20() {
        syncVMPFile();
        String ofname=getSaveFilename("Write VMP Version 2.0 file",vmpfile.dir,
                                      vmpfile.filename);
        if (ofname==null) return;

        try {
            vmpfile.writeVMP20(ofname);

            filelabel.setText(vmpfile.filename);
            dirlabel.setText(vmpfile.dir);
            currentdirectory=vmpfile.dir;

        } catch (Exception ex) {
            String text[]={"Error writing file "+ofname,""};
            String buts[]={ "Nuts" };
            text[1]=ex.getMessage();
            AreYouSure ays=new AreYouSure(this,"Error Writing VMP 2.0",
                                          text,buts);
        }
    }

    // **********************************************************************
    // getSaveFilename() -- gets a save filename, verifies that the 
    //                        directory is writeable, and verifies
    //                        that it's OK to overwrite the file
    //                        if it already exists

    private String getSaveFilename(String title,String dir,String filename) {

        FileDialog filebox=new FileDialog(this,title,FileDialog.SAVE);
        filebox.setDirectory(dir);
        filebox.setFile(filename);
        filebox.setVisible(true);

        String savedir=filebox.getDirectory();
        String savefile=filebox.getFile();
        if (savefile==null) return null;

        File odir=new File(savedir);
        File ofp=new File(savedir,savefile);

        if (!odir.canWrite()) {
            String text[]={ "Error: cannot write to directory",savedir };
            String buts[]={ "Abort Save" };
            AreYouSure ays=new AreYouSure(this,"Error Saving",text,buts);
            return null;
        }

        if (ofp.exists()) {
            String text[]={ "Overwrite file?" , savefile };
            String buts[]={ "Overwrite","Cancel" };
            AreYouSure ays=new AreYouSure(this,"Overwrite?",text,buts);
            if (!ays.result.equals("Overwrite")) return null;
        }

        return ofp.getPath();
    }

    // **********************************************************************
    // removeMap()

    public void removeMap(JVFrame jvf) {
        String text[]=new String[2];
        String buts[]={ "Remove", "Cancel" };
        text[0]="Remove map "+String.valueOf(jvf.number)+"?";
        text[1]=jvf.title;
        AreYouSure ays=new AreYouSure(this,"Remove Map",text,buts);
        if (ays.result.equals("Remove")) {
            synchronized (maplist) {
                String match=new String("Map "+
                                        String.valueOf(jvf.number)+":");
                int len=match.length();
                for (int i=0 ; i<maplist.getItemCount() ; ++i) {
                    String cur=maplist.getItem(i).substring(0,len);
                    if (cur.equals(match)) {
                        maplist.remove(i);
                        break;
                    }
                }
            }
            synchronized (maps) {
                maps.removeElement(jvf);
                jvf.cleanup();
            }
        }
    }
        

    // **********************************************************************
    // removePlayer()

    public void removePlayer(PlayerInfo pl) {
        String text[]=new String[2];
        String buts[]={ "Remove", "Cancel" };
        text[0]="Remove player "+String.valueOf(pl.number)+"?";
        text[1]=new String(pl.name+" <"+pl.email+">");
        AreYouSure ays=new AreYouSure(this,"Remove Player",text,buts);
        if (ays.result.equals("Remove")) {
            synchronized (playerlist) {
                String match=new String("Player "+
                                        String.valueOf(pl.number)+":");
                int len=match.length();
                for (int i=0 ; i<playerlist.getItemCount() ; ++i) {
                    String cur=playerlist.getItem(i).substring(0,len);
                    if (cur.equals(match)) {
                        playerlist.remove(i);
                        break;
                    }
                }
            }
            synchronized (players) {
                players.removeElement(pl);
            }
        }
    }
        

    // **********************************************************************
    // **********************************************************************
    // CSetEditor is a class that pops up a dialog and edits the counter set

    class CSetEditor extends Dialog implements ActionListener {

        private TextField imagetf,rowstf,colstf;
        private Button buttons[];
        private String errormsg;

        public CSetEditor() {
            super(JV_Map.this,"Change Counter Set",false);

            Label l;
            Panel p;
            FlowLayout flo;
            int i;
            
            buttons=new Button[3];
            errormsg=new String("");

            p=new Panel();
            flo=new FlowLayout();
            p.setLayout(flo);
            add("North",p);

            if (cset==null) {
                imagetf=new TextField("",32);
            } else {
                imagetf=new TextField(cset.filename,32);
            }
            p.add(imagetf);
            buttons[0]=new Button("Select");
            buttons[0].setActionCommand("Select Image");
            buttons[0].addActionListener(this);
            p.add(buttons[0]);
            
            p=new Panel();
            flo=new FlowLayout();
            p.setLayout(flo);
            add("Center",p);

            l=new Label("Numer of tiles: ");
            p.add(l);
            l=new Label("Rows: ");
            p.add(l);
            if (cset==null) {
                rowstf=new TextField(String.valueOf(0));
                p.add(rowstf);
            } else {
                rowstf=null;
                l=new Label(String.valueOf(cset.rows));
                p.add(l);
            }
            l=new Label("   Columns: ");
            p.add(l);
            if (cset==null) {
                colstf=new TextField(String.valueOf(0));
                p.add(colstf);
            } else {
                colstf=null;
                l=new Label(String.valueOf(cset.columns));
                p.add(l);
            }

            p=new Panel();
            flo=new FlowLayout();
            p.setLayout(flo);
            add("South",p);
            buttons[1]=new Button("OK");
            buttons[1].setActionCommand("Enact Changes");
            buttons[1].addActionListener(this);
            p.add(buttons[1]);
            buttons[2]=new Button("Cancel");
            buttons[2].setActionCommand("Cancel");
            buttons[2].addActionListener(this);
            p.add(buttons[2]);

            pack();

            // Center

            Point pnt=JV_Map.this.getLocation();
            Dimension d=JV_Map.this.getSize();
            Dimension s=getSize();
            pnt.translate((d.width-s.width)/2,(d.height-s.height)/2);
            setLocation(pnt);

            setVisible(true);
        }

        // ******************************************************************
        // ActionListener method

        public void actionPerformed(ActionEvent e) {
            String com=new String(e.getActionCommand());

            if (com.equals("Select Image")) {
                selectCsetImage();
            }
            else if (com.equals("Cancel")) {
                cleanup();
            }
            else if (com.equals("Enact Changes")) {
                if (setNewCset()) {
                    cleanup();
                }
            }
            else {
                if (debug) {
                    System.err.println("Unknown command: "+com);
                }
            }
        }

        // ******************************************************************
        // selectCsetImage()

        private void selectCsetImage() {
            FileDialog filebox=new FileDialog(JV_Map.this,"Select Cset Image",
                                              FileDialog.LOAD);
            filebox.setDirectory(vmpfile.dir);
            filebox.setVisible(true);
            String dir=filebox.getDirectory();
            String file=filebox.getFile();
            if (file==null) return;

            File newdir=new File(dir);
            File olddir=new File(vmpfile.dir);

            if (!newdir.getAbsolutePath().equals(olddir.getAbsolutePath())) {
                String text[] = {
                  "Error: all images must be in the same directory",
                  "as the VMP file" };
                String buttons[]={"Oops"};
                AreYouSure ays=new AreYouSure(this,"Counter Set Error",
                                              text,buttons);
                return;
            }
            imagetf.setText(file);
        }

        // ******************************************************************
        // setNewCset()

        private boolean setNewCset() {
            int newrows;
            int newcols;
            Image newimage;

            try {
                newimage=getImage(vmpfile.dir,imagetf.getText());
            }
            catch (Exception ex) {
                String text[]={ "Error reading image: ",""};
                String buttons[]={ "Nuts" };
                text[1]=ex.getMessage();
                AreYouSure ays=new AreYouSure(this,"Counter Set Error",
                                              text,buttons);
                return false;
            }

            if (cset!=null) {
                if (   newimage.getWidth(this)!=cset.imwidth || 
                       newimage.getHeight(this)!=cset.imheight  ) {
                    String text[]={
                        "Error: size of replacement Tiles image must",
                        "exactly match size of the current Tiles image" };
                    String buttons[]={ "Oops" };
                    AreYouSure ays=new AreYouSure(this,"Counter Set Warning",
                                                  text,buttons);
                    return false;
                }
                try {
                    cset.setImage(newimage);
                }
                catch (Exception ex) {}
            }
            else {
                try {
                    int rows,cols;
                    rows=Integer.parseInt(rowstf.getText());
                    cols=Integer.parseInt(colstf.getText());
                    cset=new CounterSet(newimage,cols,rows);
                    cset.filename=new String(imagetf.getText());
                    tilelabel.setText("Tiles image: "+cset.filename+" ("+
                                      String.valueOf(cset.columns)+" x "+
                                      String.valueOf(cset.rows)+")");
                }
                catch (Exception ex) {
                    String text[]={ "Fatal Error:",""};
                    String buts[]={ "Damn" };
                    text[1]=ex.getMessage();
                    AreYouSure ays=new AreYouSure(this,"Fatal JV_Map Error",
                                                  text,buts);
                    System.exit(0);
                }
            }
            
            return true;
        }
        
        // ******************************************************************
        // cleanup()

        private void cleanup() {
            for (int i=0 ; i<buttons.length ; ++i) {
                buttons[i].removeActionListener(this);
            }
            setVisible(false);
            dispose();
        }
    }


    // **********************************************************************
    // **********************************************************************
    // MapEditor is a class that pops up a dialog and edits a map

    class MapEditor extends Dialog implements ActionListener {

        private TextField imagetf,titletf,x0tf,y0tf,numbertf;
        private Button buttons[];

        private JVFrame inmap;

        public MapEditor(JVFrame map) {
            super(JV_Map.this,"Edit Map",false);
            if (map==null) setTitle("Add Map");

            FlowLayout flo;
            GridLayout glo;
            BorderLayout blo;
            Panel pnl;
            Panel p;
            Label l;
            int max;
            Enumeration en;
            JVFrame jvf;

            inmap=map;
            buttons=new Button[3];

            pnl=new Panel();
            glo=new GridLayout(3,1);
            pnl.setLayout(glo);
            add("Center",pnl);

            p=new Panel();
            blo=new BorderLayout();
            p.setLayout(blo);
            l=new Label("Map Image: ");
            p.add("West",l);
            imagetf=new TextField("");
            p.add("Center",imagetf);
            buttons[0]=new Button("Select");
            buttons[0].setActionCommand("Select Image");
            buttons[0].addActionListener(this);
            p.add("East",buttons[0]);
            pnl.add(p);

            p=new Panel();
            blo=new BorderLayout();
            p.setLayout(blo);
            l=new Label("Map Title: ");
            p.add("West",l);
            titletf=new TextField("");
            p.add("Center",titletf);
            pnl.add(p);

            p=new Panel();
            l=new Label("x0:");
            p.add(l);
            x0tf=new TextField("0");
            p.add(x0tf);
            l=new Label(" y0:");
            p.add(l);
            y0tf=new TextField("0");
            p.add(y0tf);
            l=new Label(" Map #:");
            p.add(l);
            if (map==null) {
                max=0;
                en=maps.elements();
                while (en.hasMoreElements()) {
                    jvf=(JVFrame)en.nextElement();
                    if (jvf.number>max) max=jvf.number;
                }
                numbertf=new TextField(String.valueOf(max+1));
                p.add(numbertf);
            } else {
                numbertf=null;
                l=new Label(String.valueOf(map.number));
                p.add(l);
            }
            pnl.add(p);


            p=new Panel();
            blo=new BorderLayout();
            p.setLayout(blo);
            if (map==null) {
                buttons[1]=new Button("Add");
            } else {
                buttons[1]=new Button("Change");
            }
            buttons[1].setActionCommand("Save Changes");
            buttons[1].addActionListener(this);
            p.add("West",buttons[1]);
            buttons[2]=new Button("Cancel");
            buttons[2].setActionCommand("Cancel");
            buttons[2].addActionListener(this);
            p.add("East",buttons[2]);
            add("South",p);

            if (map!=null) {
                imagetf.setText(map.filename);
                titletf.setText(map.title);
                x0tf.setText(String.valueOf(map.x0));
                y0tf.setText(String.valueOf(map.y0));
            }

            pack();

            // Center

            Point pnt=JV_Map.this.getLocation();
            Dimension d=JV_Map.this.getSize();
            Dimension s=getSize();
            pnt.translate((d.width-s.width)/2,(d.height-s.height)/2);
            setLocation(pnt);

            setVisible(true);
        }

        // ******************************************************************
        // ActionListener method

        public void actionPerformed(ActionEvent e) {
            String com=e.getActionCommand();

            if (com.equals("Select Image")) {
                selectMapImage();
            }
            else if (com.equals("Cancel")) {
                cleanup();
            }
            else if (com.equals("Save Changes")) {
                if (changeMap()) {
                    cleanup();
                }
            }
        }

        // ******************************************************************
        // cleanup()

        public void cleanup() {
            for (int i=0 ; i<buttons.length ; ++i) {
                buttons[i].removeActionListener(this);
            }
            setVisible(false);
            dispose();
        }

        // ******************************************************************
        // selectMapImage();
        //
        // It occurs to me that since this funciton is almost identical
        //   to selectCsetImage, I ought to make both editor classes
        //   a subclass of a common super class....

        public void selectMapImage() {
            FileDialog filebox=new FileDialog(JV_Map.this,"Select Map Image",
                                              FileDialog.LOAD);
            filebox.setDirectory(vmpfile.dir);
            filebox.setVisible(true);
            String dir=filebox.getDirectory();
            String file=filebox.getFile();
            if (file==null) return;

            File newdir=new File(dir);
            File olddir=new File(vmpfile.dir);

            if (!newdir.getAbsolutePath().equals(olddir.getAbsolutePath())) {
                String text[] = {
                  "Error: all images must be in the same directory",
                  "as the VMP file" };
                String buttons[]={"Oops"};
                AreYouSure ays=new AreYouSure(this,"Counter Set Error",
                                              text,buttons);
                return;
            }
            imagetf.setText(file);
        }

        // ******************************************************************
        // changeMap() -- save the changes, return false if it failed

        public boolean changeMap() {
            Image newimage;
            JVFrame newmap,map;
            int x0,y0,number,width,height,len;
            String title,filename,temp,cur,matchstr;
            
            filename=imagetf.getText();
            if (inmap==null || !filename.equals(inmap.filename)) {
                if (debug) {
                    System.err.println("Reading new map image.");
                }
                try {
                    newimage=getImage(vmpfile.dir,filename);
                }
                catch (Exception ex) {
                    String text[]={ "Error reading image: ",""};
                    String buttons[]={ "Nuts" };
                    text[1]=ex.getMessage();
                    AreYouSure ays=new AreYouSure(this,"Counter Set Error",
                                                  text,buttons);
                    return false;
                }
                width=newimage.getWidth(this);
                height=newimage.getHeight(this);
            }
            else {
                newimage=null;
                width=inmap.width;
                height=inmap.height;
            }                

            if (inmap!=null) {
                if (newimage!=null) {
                    if (width!=inmap.width || height!=inmap.height) {
                        String text[]={
                            "Error: size of replacement Map image must",
                            "exactly match size of the current Map image" };
                        String buttons[]={ "Oops" };
                        AreYouSure ays=new AreYouSure(this,"Map Edit Warning",
                                                      text,buttons);
                        return false;
                    }
                }
                number=inmap.number;
            } else {
                number=Integer.parseInt(numbertf.getText());
                Enumeration en=maps.elements();
                while (en.hasMoreElements()) {
                    map=(JVFrame)en.nextElement();
                    if (number==map.number) {
                        String text[]={ "Error: map "+String.valueOf(number)+
                                        " exists.",
                                        "Map numbers must be unique" };
                        String buts[]={ "Oh" };
                        AreYouSure ays=new AreYouSure(this,"Add Map Warning",
                                                      text,buts);
                        return false;
                    }
                }
            }

            x0=Integer.parseInt(x0tf.getText());
            y0=Integer.parseInt(x0tf.getText());
            title=titletf.getText();

            if (inmap!=null) {

                if (newimage!=null) {
                    inmap.getCanvas().setImage(newimage);
                }
                inmap.x0=x0;
                inmap.y0=y0;
                inmap.title=new String(title);
                inmap.filename=new String(filename);
                inmap.setTitle("JV_Map "+jvmapversion+": "+title);

                // Alter the entry in the maplist widget

                temp=new String("Map "+String.valueOf(inmap.number)+
                                ": "+inmap.title+" ("+inmap.filename+")");

                matchstr=new String("Map "+String.valueOf(inmap.number)+":");
                len=matchstr.length();

                synchronized (maplist) {
                    for (int i=0 ; i<maplist.getItemCount() ; ++i) {
                        cur=maplist.getItem(i);
                        if (cur.substring(0,len).equals(matchstr)) {
                            maplist.replaceItem(temp,i);
                            break;
                        }
                    }
                }
            }
            else {

                // Add the new map

                try {
                    newmap=new JVFrame(newimage,cset,number,x0,y0,width,height,
                                       filename,title);
                }
                catch (Exception ex) {
                    String text[]={ "Error making map:",""};
                    String buts[]={ "Nuts" };
                    text[1]=ex.getMessage();
                    AreYouSure ays=new AreYouSure(this,"Add Map Error",
                                                  text,buts);
                    return false;
                }
                newmap.setTitle("JV_Map "+jvmapversion+": "+title);

                maps.addElement(newmap);

                // Add to the map list gadget
                temp=new String("Map "+String.valueOf(newmap.number)+
                                ": "+newmap.title+" ("+newmap.filename+")");
                maplist.add(temp);
            }

            return true;
        }

    }
    
    // **********************************************************************
    // **********************************************************************
    // TitleEditor

    class TitleEditor extends Dialog implements ActionListener {

        private TextField titletf;
        private Button okbutton,cancelbutton;

        public TitleEditor() {
            super(JV_Map.this,"Edit Game Title",false);

            BorderLayout blo;
            Panel p;
            Label l;

            p=new Panel();
            blo=new BorderLayout();
            p.setLayout(blo);
            l=new Label("Game Title: ");
            p.add("West",l);
            titletf=new TextField(gametitle,32);
            titletf.addActionListener(this);
            p.add("Center",titletf);
            add("Center",p);

            p=new Panel();
            blo=new BorderLayout();
            p.setLayout(blo);
            okbutton=new Button("OK");
            okbutton.setActionCommand("OK");
            okbutton.addActionListener(this);
            p.add("West",okbutton);
            cancelbutton=new Button("Cancel");
            cancelbutton.setActionCommand("Cancel");
            cancelbutton.addActionListener(this);
            p.add("East",cancelbutton);
            add("South",p);

            pack();

            // Center

            Point pnt=JV_Map.this.getLocation();
            Dimension d=JV_Map.this.getSize();
            Dimension s=getSize();
            pnt.translate((d.width-s.width)/2,(d.height-s.height)/2);
            setLocation(pnt);

            setVisible(true);
        }

        // ******************************************************************
        // cleanup()

        public void cleanup() {
            titletf.removeActionListener(this);
            okbutton.removeActionListener(this);
            cancelbutton.removeActionListener(this);
            dispose();
        }

        // ******************************************************************
        // actionPerformed()

        public void actionPerformed(ActionEvent e) {
            if (!e.getActionCommand().equals("Cancel")) {
                gametitle=new String(titletf.getText());
                titlelabel.setText(gametitle);
                JV_Map.this.setTitle("JV_Map "+jvmapversion+": "+gametitle);
            }
            cleanup();
        }
    }
        

    // **********************************************************************
    // **********************************************************************
    // PlayerEditor() does what you think it does

    class PlayerEditor extends Dialog implements ActionListener {
        
        private TextField nametf,numbertf,emailtf;
        private Button changebutton,cancelbutton;

        private PlayerInfo inplayer;

        public PlayerEditor(PlayerInfo player) {
            super(JV_Map.this,"Edit Player",false);
            if (player==null) setTitle("Add Player");

            Panel p1,p2,p3;
            BorderLayout blo;
            Label l;
            int i,max;
            Enumeration en;
            PlayerInfo pl;

            inplayer=player;

            p1=new Panel();
            add("North",p1);
            p2=new Panel();
            blo=new BorderLayout();
            p2.setLayout(blo);
            p1.add("North",p2);
            p3=new Panel();
            p2.add("North",p3);
            l=new Label("Player Number: ");
            p3.add(l);
            if (player==null) {
                max=0;
                en=players.elements();
                while (en.hasMoreElements()) {
                    pl=(PlayerInfo)en.nextElement();
                    if (pl.number>max) max=pl.number;
                }
                numbertf=new TextField(String.valueOf(max+1));
                p3.add(numbertf);
                p3=new Panel();
                p2.add("Center",p3);
                l=new Label("Note: Player 0 should be reserved for the GM");
                p3.add(l);
            } else {
                l=new Label(String.valueOf(player.number));
                p3.add(l);
            }

            p1=new Panel();
            blo=new BorderLayout();
            p1.setLayout(blo);
            add("Center",p1);

            p2=new Panel();
            blo=new BorderLayout();
            p2.setLayout(blo);
            p1.add("North",p2);
            l=new Label("Name:");
            p2.add("West",l);
            if (player==null) {
                nametf=new TextField("");
            } else {
                nametf=new TextField(player.name);
            }
            p2.add("Center",nametf);

            p2=new Panel();
            blo=new BorderLayout();
            p2.setLayout(blo);
            p1.add("Center",p2);
            l=new Label("Email:");
            p2.add("West",l);
            if (player==null) {
                emailtf=new TextField("");
            } else {
                emailtf=new TextField(player.email);
            }
            p2.add("Center",emailtf);


            p1=new Panel();
            blo=new BorderLayout();
            p1.setLayout(blo);
            add("South",p1);
            if (player==null) {
                changebutton=new Button("Add");
            } else {
                changebutton=new Button("Change");
            }
            changebutton.setActionCommand("Change");
            changebutton.addActionListener(this);
            p1.add("West",changebutton);

            cancelbutton=new Button("Cancel");
            cancelbutton.setActionCommand("Cancel");
            cancelbutton.addActionListener(this);
            p1.add("East",cancelbutton);

            pack();

            // Center

            Point pnt=JV_Map.this.getLocation();
            Dimension d=JV_Map.this.getSize();
            Dimension s=getSize();
            pnt.translate((d.width-s.width)/2,(d.height-s.height)/2);
            setLocation(pnt);

            setVisible(true);
        }
            
        // ******************************************************************
        // cleanup()

        public void cleanup() {
            setVisible(false);
            cancelbutton.removeActionListener(this);
            changebutton.removeActionListener(this);
            dispose();
        }

        // ******************************************************************
        // actionPerformed()

        public void actionPerformed(ActionEvent e) {
            String com=e.getActionCommand();
            String newstr,matchstr,cur;
            int matchlen,i,num;
            PlayerInfo pl;
            
            if (com.equals("Cancel")) cleanup();
            if (!com.equals("Change")) return;

            if (inplayer==null) {
                num=Integer.parseInt(numbertf.getText());
            } else {
                num=inplayer.number;
            }
            newstr=new String("Player "+String.valueOf(num)+": "+
                              nametf.getText()+" <"+
                              emailtf.getText()+">");

            if (inplayer!=null) {
                matchstr=new String("Player "+String.valueOf(inplayer.number)+
                                    ":");
                matchlen=matchstr.length();
                synchronized (playerlist) {
                    for (i=0 ; i<playerlist.getItemCount() ; ++i) {
                        cur=playerlist.getItem(i);
                        if (cur.substring(0,matchlen).equals(matchstr)) {
                            playerlist.replaceItem(newstr,i);
                            break;
                        }
                    }
                }
                inplayer.name=new String(nametf.getText());
                inplayer.email=new String(emailtf.getText());
            } else {
                synchronized (players) {
                    Enumeration en=players.elements();
                    while (en.hasMoreElements()) {
                        pl=(PlayerInfo)en.nextElement();
                        if (pl.number==num) {
                            String text[]={ "Error: player numbers must "+
                                            "be unique" };
                            String buts[]={ "Oh" };
                            AreYouSure ays=new AreYouSure(this,"Player Error",
                                                          text,buts);
                            return;
                        }
                    }
                }
                pl=new PlayerInfo();
                pl.number=num;
                pl.email=new String(emailtf.getText());
                pl.name=new String(nametf.getText());
                pl.password=new String("*");
                players.addElement(pl);
                playerlist.add(newstr);
            }
            cleanup();
        }
    }
            


    // **********************************************************************
    // **********************************************************************
    // AboutBox() is an asynchronous Dialog Box with some additional info

    class AboutBox implements ActionListener {
        FlowLayout flow;
        Button b;
        Dialog dlg;

        public void addLabel(Panel panel,String text,Font font) {
            Label l;
            Panel subpanel;

            subpanel=new Panel();
            subpanel.setLayout(flow);
            l=new Label(text);
            l.setFont(font);
            subpanel.add(l);
            panel.add(subpanel);
        }      

        public AboutBox(Frame frame) {
            dlg=new Dialog(frame,"About JV_Map",false);
            GridLayout grid=new GridLayout(5,1);         // GridBag it!!
            Panel panel=new Panel();
            Font bigSerif=new Font("Serif",Font.BOLD,24);
            Font normal=new Font("SansSerif",0,12);

            flow=new FlowLayout(FlowLayout.CENTER);

            panel.setLayout(grid);

            addLabel(panel,"JV_Map",bigSerif);
            addLabel(panel,"Version "+jvmapversion,normal);
            addLabel(panel,"A Java wargame PBEM aid compatable with V_MAP",
                     normal);
            addLabel(panel,"by Robert Knop (rknop@pobox.com)",normal);
            addLabel(panel,"http://www.pobox.com/~rknop/jv_map",normal);

            dlg.add("Center",panel);

            b=new Button("Dismiss");
            b.addActionListener(this);
            dlg.add("South",b);
            dlg.pack();

            // Center

            Point pnt=JV_Map.this.getLocation();
            Dimension d=JV_Map.this.getSize();
            Dimension s=dlg.getSize();
            pnt.translate((d.width-s.width)/2,(d.height-s.height)/2);
            dlg.setLocation(pnt);

            dlg.show();
        }

        public void cleanup() {
            dlg.removeAll();
            dlg.setVisible(false);
            dlg.dispose();
            b.removeActionListener(this);
        }

        public void actionPerformed(ActionEvent e) {
            cleanup();
        }

        public void finalize() throws Throwable {
            if (debug) {
                System.out.println("In AboutBox finalizer.");
            }
            super.finalize();
        }
    }

    // **********************************************************************
    // **********************************************************************
    // main() -- on startup, find out if a string was specified.  If so,
    //   assume that that's the VMAP file, and read it.
    
    public static void main(String args[]) {
        JV_Map map;
        VMPFile vmp;
        
        try {
            if (args.length==0) {
                map=new JV_Map();
            }
            else {
                map=new JV_Map(args[0]);
                map.setVisible(true);
            }
        }
        catch (Exception ex) {
            System.err.println(ex.getMessage());
            System.exit(20);
        }
    }

}
